//------------------------------------------------------------------------------
//  Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
// 
//  The contents of this file are subject to the Blocks Public License (the 
//  "License"); you may not use this file except in compliance with the License.  
//  You may obtain a copy of the License at http://www.invisible.net/ 
// 
//  Software distributed under the License is distributed on an "AS IS" basis,  
//  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for  
//  the specific language governing rights and limitations under the License. 
//------------------------------------------------------------------------------
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

public class InvisibleOtp {

byte []                                 hash;

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public InvisibleOtp() {
     }

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public InvisibleOtp(String algorithm,String phrase,String seed,int count) throws InvisibleException {
     if((phrase == null) || (phrase.length() == 0)) {
          throw new InvisibleException("InvisibleOtp: Missing Pass Phrase");
          }

     if((seed == null) || (seed.length() == 0)) {
          throw new InvisibleException("InvisibleOtp: Missing Seed");
          }

     algorithm = make_algorithm(algorithm);

     try {
          hash = make_hash(algorithm,make_phrase(seed,phrase));
          hash = make_folded(algorithm,hash);

          for(int k=0;k<count;++k) {
               hash = make_hash(algorithm,hash);
               hash = make_folded(algorithm,hash);
               }
          }
     catch(NoSuchAlgorithmException e) {
          throw new InvisibleException("InvisibleOtp: Unknown Algorithm: " + algorithm);
          }
     }

//------------------------------------------------------------------------------
//  set hash result
//------------------------------------------------------------------------------
public void set(String s) throws InvisibleException {
     //-------------------------------------------------------------------------
     //  set from hex format
     //-------------------------------------------------------------------------
     s = s.toUpperCase();

     if(s.startsWith("HEX:")) {
          set_hash(s.substring(4));
          return;
          }

     //-------------------------------------------------------------------------
     //  set from word format
     //-------------------------------------------------------------------------
     if(s.startsWith("WORD:")) {
          InvisibleVector v = InvisibleUtil.parse(s.substring(5)," ");

          if(v.size() != 6) {
               throw new InvisibleException("InvisibleOtp: Wrong Number Of Words: " + s);
               }

          long newHash = 0L;

          for(int k=0;k<6;++k) {
               String word = v.getString(k);

               int index = InvisibleOtpDictionary.find(word);

               if(k == 5) {
                    newHash = (newHash << 9) | (index >> 2);
                    }
               else {
                    newHash = (newHash << 11) | index;
                    }
               }

          set_hash(newHash);

          return;
          }

     //-------------------------------------------------------------------------
     //  unsupported format
     //-------------------------------------------------------------------------
     throw new InvisibleException("InvisibleOtp: Unsupported Format: " + s);
     }

//------------------------------------------------------------------------------
//  set hash result from string
//------------------------------------------------------------------------------
private void set_hash(String s) {
     long l = Long.valueOf(s,16).longValue();
     set_hash(l);
     }

//------------------------------------------------------------------------------
//  set hash result from long
//------------------------------------------------------------------------------
private void set_hash(long l) {
     hash = new byte[8];

     for(int k=7;k>=0;--k) {
          hash[k] = (byte)(l & 0xff);
          l >>= 8;
          }
     }

//------------------------------------------------------------------------------
//  get hash result in specified format
//------------------------------------------------------------------------------
public String get(String format) throws InvisibleException {
     //-------------------------------------------------------------------------
     //  hex
     //-------------------------------------------------------------------------
     if(format.equalsIgnoreCase("hex")) {
          return("hex:"+make_hex());
          }

     //-------------------------------------------------------------------------
     //  word
     //-------------------------------------------------------------------------
     if(format.equalsIgnoreCase("word")) {
          long l = make_long();
          StringBuffer sb = new StringBuffer("word:");

          sb.append(InvisibleOtpDictionary.get(make_index(l,1)));
          sb.append(" ");
          sb.append(InvisibleOtpDictionary.get(make_index(l,2)));
          sb.append(" ");
          sb.append(InvisibleOtpDictionary.get(make_index(l,3)));
          sb.append(" ");
          sb.append(InvisibleOtpDictionary.get(make_index(l,4)));
          sb.append(" ");
          sb.append(InvisibleOtpDictionary.get(make_index(l,5)));
          sb.append(" ");
          sb.append(InvisibleOtpDictionary.get(make_index(l,6)));

          return(sb.toString());
          }

     throw new InvisibleException("InvisibleOtp: Unknown Format");
     }

//------------------------------------------------------------------------------
//  make an 11-bit index for word lookup
//------------------------------------------------------------------------------
private int make_index(long l,int num) throws InvisibleException {
     switch(num) {
          case 1: return((int)((l >> 53) & 0x7ff));
          case 2: return((int)((l >> 42) & 0x7ff));
          case 3: return((int)((l >> 31) & 0x7ff));
          case 4: return((int)((l >> 20) & 0x7ff));
          case 5: return((int)((l >> 9) & 0x7ff));
          case 6:
               long tmp = l;
               byte parity = 0;
               for(int k=0;k<64;k+=2) {
                    parity += tmp & 0x03;
                    tmp >>= 2;
                    }
               return((int)(((((l << 2) & 0x7fc) | (parity & 0x03))) & 0x7ff));
          }

     throw new InvisibleException("InvisibleOtp: Word Index Out-Of-Range");
     }

//------------------------------------------------------------------------------
//  make a hex string from the hash result
//------------------------------------------------------------------------------
private String make_hex() {
     StringBuffer sb = new StringBuffer();
     for(int k=0;k<8;++k) {
          String s =  Integer.toHexString(0xff & hash[k]);
          if(s.length() < 2) sb.append("0");
          sb.append(s);
          }

     return(sb.toString());
     }

//------------------------------------------------------------------------------
//  make a long from the hash result
//------------------------------------------------------------------------------
private long make_long() {
     long l = 0L;

     for(int k=0;k<8;++k) {
          l = (l << 8) | (hash[k] & 0xff);
          }

     return(l);
     }

//------------------------------------------------------------------------------
//  hash the pharse with the supplied algorithm
//------------------------------------------------------------------------------
private byte[] make_hash(String algorithm,byte[] phrase) throws NoSuchAlgorithmException {
     MessageDigest digest = MessageDigest.getInstance(algorithm);
     digest.reset();
     digest.update(phrase);
     return(digest.digest());
     }

//------------------------------------------------------------------------------
//  convert hash results to little-endian format
//------------------------------------------------------------------------------
private byte[] make_littleendian(byte[] phrase) {
     byte [] newPhrase = new byte[8];

     for(int k=0;k<4;++k) {
          newPhrase[3-k] = phrase[k];
          newPhrase[7-k] = phrase[k+4];
          }

     return(newPhrase);
     }

//------------------------------------------------------------------------------
//  fold the has results to 64 bits
//------------------------------------------------------------------------------
private byte[] make_folded(String algorithm,byte[] phrase) {
     if(algorithm.equalsIgnoreCase("MD5")) {
          for(int k=0;k<8;++k) phrase[k] ^= (0xff & phrase[k+8]);
          }
     else {
          if(algorithm.equalsIgnoreCase("SHA")) {
               for(int k=0;k<4;++k) phrase[k] ^= (0xff & phrase[k+8]);
               for(int k=4;k<8;++k) phrase[k] ^= (0xff & phrase[k+8]);
               for(int k=0;k<4;++k) phrase[k] ^= (0xff & phrase[k+16]);
               return(make_littleendian(phrase));
               }
          }

     byte [] newPhrase = new byte[8];
     for(int k=0;k<8;++k) newPhrase[k] = phrase[k];

     return(newPhrase);
     }

//------------------------------------------------------------------------------
//  create a hash result from the supplied seed and phrase string
//------------------------------------------------------------------------------
private byte[] make_phrase(String seed,String phrase) throws InvisibleException {
     if(phrase.length() < 10) {
          throw new InvisibleException("InvisibleOtp: Pass Phrase Too Short");
          }
     String combined = seed+phrase;
     return(combined.getBytes());
     }

//------------------------------------------------------------------------------
//  make algorithm name
//------------------------------------------------------------------------------
private String make_algorithm(String s) {
     if(s.equalsIgnoreCase("otp-sha1")) return("SHA");
     if(s.equalsIgnoreCase("otp-sha")) return("SHA");
     if(s.equalsIgnoreCase("sha1")) return("SHA");
     if(s.equalsIgnoreCase("otp-md5")) return("MD5");
     return(s);
     }
}
