//------------------------------------------------------------------------------
//  Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
// 
//  The contents of this file are subject to the Blocks Public License (the 
//  "License"); you may not use this file except in compliance with the License.  
//  You may obtain a copy of the License at http://www.invisible.net/ 
// 
//  Software distributed under the License is distributed on an "AS IS" basis,  
//  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for  
//  the specific language governing rights and limitations under the License. 
//------------------------------------------------------------------------------
import java.net.InetAddress;

public class SpaceServer {

public final static int                 defaultPort = 10288;

private SpaceServerThread               thread;
private SpaceSocket                     socket;
private SpaceChannel                    channelZero;
private SpaceResponse                   channelZeroResponse;
private SpaceCache                      cache;
private SpaceChannelCache               requestChannels;
private SpaceRequestQueue               requestQueue;
private SpaceSerialCache                serialCache;
private SpaceLog                        log;
private String[]                        profiles;
private String                          localize;
private String                          lastLocalize;
private String                          features;
private String                          authenticator;
private int                             nextChannel;
private int                             reqno;
private boolean                         trace;
private boolean                         debug;

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceServer() throws SpaceException {
     init(true,defaultPort);
     }

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceServer(int port) throws SpaceException {
     init(true,port);
     }

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceServer(String host) throws SpaceException {
     init(true,defaultPort);
     setHost(host);
     }

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceServer(String host,int port) throws SpaceException {
     init(true,port);
     setHost(host);
     setPort(port);
     }

//------------------------------------------------------------------------------
//  init
//------------------------------------------------------------------------------
private void init(boolean newconnection,int port) throws SpaceException {
     if(newconnection) socket = new SpaceSocket(port);
     channelZero = getChannel(0,null);
     serialCache = new SpaceSerialCache();
     log = new SpaceLog();
     cache = new SpaceCache();
     channelZeroResponse = new SpaceResponse(channelZero,0);
     cache.set(0,channelZeroResponse);
     requestChannels = new SpaceChannelCache();
     requestQueue = new SpaceRequestQueue();
     requestChannels.set(0,channelZero);
     profiles = null;
     localize = null;
     features = null;
     authenticator = null;
     nextChannel = -1; // +2 == 1
     if(newconnection) reqno = 1;
     if(newconnection) thread = null;
     setDebug(debug);
     }

//------------------------------------------------------------------------------
//  get next channel
//------------------------------------------------------------------------------
synchronized public SpaceChannel getChannel(String profile) throws SpaceException {
     nextChannel += 2;
     return(getChannel(nextChannel,profile));
     }

//------------------------------------------------------------------------------
//  get channel
//------------------------------------------------------------------------------
private SpaceChannel getChannel(int channel,String profile) throws SpaceException {
     SpaceChannel channelElement = new SpaceChannel(channel);
     channelElement.setProfile(profile);
     channelElement.setServer(this);
     return(channelElement);
     }

//------------------------------------------------------------------------------
//  get an open channel if it exists
//------------------------------------------------------------------------------
public SpaceChannel getChannel(int channel) throws SpaceException {
     return(requestChannels.get(channel));
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
public SpaceResponse connect() throws SpaceTimeout,SpaceException {
     return(connect(true,SpaceSocket.defaultTimeout));
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
public SpaceResponse connect(boolean wait) throws SpaceTimeout,SpaceException {
     return(connect(wait,SpaceSocket.defaultTimeout));
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
private SpaceResponse connect(boolean wait,int timeout) throws SpaceTimeout,SpaceException {
     return(connect(null,wait,timeout,true,false));
     }

//------------------------------------------------------------------------------
//  connect with localization
//------------------------------------------------------------------------------
public SpaceResponse connect(String specifyLocalize) throws SpaceTimeout,SpaceException {
     return(connect(specifyLocalize,true,SpaceSocket.defaultTimeout));
     }

//------------------------------------------------------------------------------
//  connect with localization
//------------------------------------------------------------------------------
public SpaceResponse connect(String specifyLocalize,boolean wait) throws SpaceTimeout,SpaceException {
     return(connect(specifyLocalize,wait,SpaceSocket.defaultTimeout));
     }

//------------------------------------------------------------------------------
//  connect with localization
//------------------------------------------------------------------------------
private SpaceResponse connect(String specifyLocalize,boolean wait,int timeout) throws SpaceTimeout,SpaceException {
     return(connect(specifyLocalize,wait,timeout,true,false));
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
private SpaceResponse connect(String specifyLocalize,boolean wait,int timeout,boolean startup,boolean reconnect) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  connect without wait
     //-------------------------------------------------------------------------
     SpaceResponse response = connect(timeout,startup,reconnect);

     //-------------------------------------------------------------------------
     //  greet
     //-------------------------------------------------------------------------
     greet(specifyLocalize);

     //-------------------------------------------------------------------------
     //  wait for the complete response if appropriate
     //-------------------------------------------------------------------------
     if(wait) {
          response = wait(response);

          //--------------------------------------------------------------------
          //  parse the greeting
          //--------------------------------------------------------------------
          parseGreeting(response);
          }

     return(response);
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
private SpaceResponse connect(int timeout,boolean startup,boolean reconnect) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  init if attempting to reconnect
     //-------------------------------------------------------------------------
     if(reconnect) {
          init(false,-1);
          }

     //-------------------------------------------------------------------------
     //  connect on the socket
     //-------------------------------------------------------------------------
     if(socket == null) init(true,defaultPort);
     if(timeout != -1) socket.setTimeout(timeout);
     socket.connect();

     //-------------------------------------------------------------------------
     //  instantiate the greeting response
     //-------------------------------------------------------------------------
     SpaceResponse response = new SpaceResponse(channelZero,0);

     cache.set(0,response);

     //-------------------------------------------------------------------------
     //  start reading responses
     //-------------------------------------------------------------------------
     if(startup) {
          thread = new SpaceServerThread(true);
          }

     thread.setSerialCache(serialCache);
     thread.setRequestChannels(requestChannels);
     thread.setRequestQueue(requestQueue);
     thread.setSocket(socket);
     thread.setCache(cache);
     thread.setLog(log);
     thread.setTrace(trace);

     if(startup) {
          thread.start();
          }

     //-------------------------------------------------------------------------
     //  notify for response loop timeout wakeup
     //-------------------------------------------------------------------------
     thread.threadNotifyAll();

     return(response);
     }

//------------------------------------------------------------------------------
//  client greeting (default localize)
//------------------------------------------------------------------------------
public void greet() throws SpaceTimeout,SpaceException {
     greet(getLocalize());
     }

//------------------------------------------------------------------------------
//  client greeting
//------------------------------------------------------------------------------
public void greet(String specifyLocalize) throws SpaceTimeout,SpaceException {
     StringBuffer sb = new StringBuffer();
     sb.append("<greeting localize='");
     if((specifyLocalize == null) || (specifyLocalize.length() == 0)) {
          sb.append(getLocalize());
          }
     else {
          sb.append(specifyLocalize);
          }
     sb.append("'/>");

     rsp(channelZero,sb.toString(),0);
     }

//------------------------------------------------------------------------------
//  parse a greeting
//------------------------------------------------------------------------------
public void parseGreeting(SpaceResponse response) throws SpaceException {
     response.parse();
     profiles = response.getProfiles();
     localize = response.getLocalize();
     features = response.getFeatures();
     }

//------------------------------------------------------------------------------
//  acknowledge a server start channel
//------------------------------------------------------------------------------
public SpaceChannel start(int channel,String profile,int serial) throws SpaceTimeout,SpaceException {
     SpaceChannel channelElement = requestChannels.set(channel,profile);

     StringBuffer sb = new StringBuffer();
     sb.append("<profile uri='");
     sb.append(profile);
     sb.append("' />");

     rsp(channelElement,sb.toString(),serial);

     return(channelElement);
     }

//------------------------------------------------------------------------------
//  acknowledge a server start channel with payload
//------------------------------------------------------------------------------
public SpaceChannel start(int channel,String profile,int serial,String payload) throws SpaceTimeout,SpaceException {
     SpaceChannel channelElement = requestChannels.set(channel,profile);

     StringBuffer sb = new StringBuffer();
     sb.append("<profile uri='");
     sb.append(profile);
     sb.append("'>");
     sb.append(payload);
     sb.append("</profile>");

     rsp(channelElement,sb.toString(),serial);

     return(channelElement);
     }

//------------------------------------------------------------------------------
//  start a channel with wait
//------------------------------------------------------------------------------
public SpaceResponse start(SpaceChannel channelElement) throws SpaceTimeout,SpaceException {
     return(start(channelElement,true));
     }

//------------------------------------------------------------------------------
//  start a channel with profile
//------------------------------------------------------------------------------
public SpaceResponse start(SpaceChannel channelElement,boolean wait) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  mark channel started
     //-------------------------------------------------------------------------
     channelElement.setStarted(true);

     //-------------------------------------------------------------------------
     //  get start response
     //-------------------------------------------------------------------------
     SpaceResponse response = request(channelZero,channelElement,true,false,null,wait,true);

     //-------------------------------------------------------------------------
     //  parse complete responses
     //-------------------------------------------------------------------------
     if(wait) {
          response.parse();
          }

     return(response);
     }

//------------------------------------------------------------------------------
//  error
//------------------------------------------------------------------------------
public void error(SpaceChannel channelElement,int messageCode,String errorString,int serial) throws SpaceTimeout,SpaceException {
     error(channelElement,messageCode,errorString,serial,getLocalize());
     }

//------------------------------------------------------------------------------
//  error
//------------------------------------------------------------------------------
public void error(SpaceChannel channelElement,int messageCode,int serial) throws SpaceTimeout,SpaceException {
     error(channelElement,messageCode,getMessage(messageCode),serial,lastLocalize);
     }

//------------------------------------------------------------------------------
//  error
//------------------------------------------------------------------------------
private void error(SpaceChannel channelElement,int messageCode,String errorString,int serial,String specifyLocalize) throws SpaceTimeout,SpaceException {
     StringBuffer sb = new StringBuffer();
     sb.append("<error code='");
     sb.append(messageCode);
     sb.append("' xml:lang='");
     sb.append(specifyLocalize);
     sb.append("'>");
     sb.append(errorString);
     sb.append("</error>");

     rsp(channelElement,sb.toString(),serial);
     }

//------------------------------------------------------------------------------
//  get authenticated channel
//------------------------------------------------------------------------------
public SpaceChannel getChannel(SpaceAuthenticator authenticator) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  start a channel for authentication
     //-------------------------------------------------------------------------
     SpaceChannel channel = getChannel(authenticator.getProfile());

     //-------------------------------------------------------------------------
     //  mark channel started
     //-------------------------------------------------------------------------
     channel.setStarted(true);

     //-------------------------------------------------------------------------
     //  request authentication
     //-------------------------------------------------------------------------
     SpaceResponse response = request(channelZero,channel,true,false,authenticator.getRequest(),true,false);

     response.parse();

     //-------------------------------------------------------------------------
     //  see if error
     //-------------------------------------------------------------------------
     if(response.isError()) {
          //--------------------------------------------------------------------
          //  error
          //--------------------------------------------------------------------
          channel.setError(response);
          }
     else {
          //--------------------------------------------------------------------
          //  see if challenge
          //--------------------------------------------------------------------
          if(response.isChallenge()) {
               //---------------------------------------------------------------
               //  respond to the challenge
               //---------------------------------------------------------------
               response = respond(channel,authenticator.getResponse(response));

               response.parse();

               if(response.isError()) {
                    //----------------------------------------------------------
                    //  unable to logon
                    //----------------------------------------------------------
                    channel.setError(response);
                    }
               }
          }

     return(channel);
     }

//------------------------------------------------------------------------------
//  respond with wait
//------------------------------------------------------------------------------
public SpaceResponse respond(SpaceChannel channelElement,String request) throws SpaceTimeout,SpaceException {
     return(respond(channelElement,request,true));
     }

//------------------------------------------------------------------------------
//  respond
//------------------------------------------------------------------------------
public SpaceResponse respond(SpaceChannel channelElement,String request,boolean wait) throws SpaceTimeout,SpaceException {
     return(request(channelElement,null,false,false,request,wait,false));
     }

//------------------------------------------------------------------------------
//  request with wait
//------------------------------------------------------------------------------
public SpaceResponse request(SpaceChannel channelElement,String request) throws SpaceTimeout,SpaceException {
     return(request(channelElement,request,true));
     }

//------------------------------------------------------------------------------
//  request
//------------------------------------------------------------------------------
public SpaceResponse request(SpaceChannel channelElement,String request,boolean wait) throws SpaceTimeout,SpaceException {
     return(request(channelElement,null,false,true,request,wait,false));
     }

//------------------------------------------------------------------------------
//  request
//------------------------------------------------------------------------------
private SpaceResponse request(SpaceChannel channelElement,SpaceChannel channelElementStart,boolean includeProfile,boolean includeReqno,String request,boolean wait,boolean createChannel) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  build request payload
     //-------------------------------------------------------------------------
     StringBuffer payload = new StringBuffer();

     //-------------------------------------------------------------------------
     //  include channel number if appropriate
     //-------------------------------------------------------------------------
     if(channelElementStart != null) {
          //--------------------------------------------------------------------
          //  make sure starting channel valid
          //--------------------------------------------------------------------
          if((channelElementStart.getChannel() & 1) != 1) {
               throw new SpaceException("SpaceChannel: Channel Must Be Odd Number");
               }

          //--------------------------------------------------------------------
          //  mark the channel started
          //--------------------------------------------------------------------
          payload.append("<start number='");
          payload.append(channelElementStart.getChannel());
          payload.append("'>\r\n");
          }

     //-------------------------------------------------------------------------
     //  include profile if appropriate
     //-------------------------------------------------------------------------
     String profile = null;

     if((includeProfile) && (channelElementStart != null) && ((profile = channelElementStart.getProfile()) != null)) {
          //--------------------------------------------------------------------
          //  include profile
          //--------------------------------------------------------------------
          payload.append("<profile uri='");
          payload.append(profile);
          payload.append("'>\r\n");

          includeProfile = true;
          }
     else {
          includeProfile = false;
          }

     //-------------------------------------------------------------------------
     //  include reqno if appropriate
     //-------------------------------------------------------------------------
     int thisReqno = -1;

     if(includeReqno) {
          thisReqno = reqno;

          payload.append("<request reqno='");
          payload.append(reqno++);
          payload.append("'>\r\n");
          }

     //-------------------------------------------------------------------------
     //  request
     //-------------------------------------------------------------------------
     if(request != null) {
          payload.append(request);
          }

     //-------------------------------------------------------------------------
     //  close reqno if appropriate
     //-------------------------------------------------------------------------
     if(includeReqno) {
          payload.append("</request>\r\n");
          }

     //-------------------------------------------------------------------------
     //  close profile if appropriate
     //-------------------------------------------------------------------------
     if(includeProfile) {
          payload.append("</profile>\r\n");
          }

     //-------------------------------------------------------------------------
     //  close channel start if appropriate
     //-------------------------------------------------------------------------
     if(channelElementStart != null) {
          payload.append("</start>\r\n");
          }

     //-------------------------------------------------------------------------
     //  make the request
     //-------------------------------------------------------------------------
     SpaceResponse response = req(channelElement,payload.toString());

     //-------------------------------------------------------------------------
     //  set the reqno
     //-------------------------------------------------------------------------
     response.setReqno(thisReqno);

     //-------------------------------------------------------------------------
     //  wait for the complete response if appropriate
     //-------------------------------------------------------------------------
     if(wait) {
          response = wait(response);
          }

     return(response);
     }

//------------------------------------------------------------------------------
//  request without auto connect
//------------------------------------------------------------------------------
private SpaceResponse req(SpaceChannel channelElement,String payload) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  build request
     //-------------------------------------------------------------------------
     StringBuffer request = new StringBuffer();

     request.append("REQ ");
     request.append(". ");
     int serial = serialCache.get();
     request.append(String.valueOf(serial));
     request.append(" ");
     request.append(String.valueOf(channelElement.getSeqno()));
     request.append(" ");
     request.append(payload.length());
     request.append(" ");
     request.append(String.valueOf(channelElement.getChannel()));
     request.append("\r\n");
     String mime = channelElement.getMime();
     if(mime != null) request.append(mime);
     request.append("\r\n");
     request.append(payload);
     request.append("END\r\n");
 
     //-------------------------------------------------------------------------
     //  write request
     //-------------------------------------------------------------------------
     socket.write(request.toString());

     //-------------------------------------------------------------------------
     //  update seqno for the channel
     //-------------------------------------------------------------------------
     channelElement.setNextSeqno(payload.length());

     //-------------------------------------------------------------------------
     //  instantiate a response for this serial
     //-------------------------------------------------------------------------
     SpaceResponse response = new SpaceResponse(channelElement,serial);

     //-------------------------------------------------------------------------
     //  set new response in the cache
     //-------------------------------------------------------------------------
     cache.set(serial,response);

     return(response);
     }

//------------------------------------------------------------------------------
//  accept
//------------------------------------------------------------------------------
public void accept(boolean useStdio) throws SpaceException {
     socket.accept(useStdio);

     thread = new SpaceServerThread(false);

     thread.setSerialCache(serialCache);
     thread.setRequestChannels(requestChannels);
     thread.setRequestQueue(requestQueue);
     thread.setSocket(socket);
     thread.setCache(cache);
     thread.setTrace(trace);

     thread.start();
     thread.threadNotifyAll();
     }

//------------------------------------------------------------------------------
//  response
//------------------------------------------------------------------------------
public void rsp(SpaceChannel channelElement,String payload) throws SpaceTimeout,SpaceException {
     rsp(channelElement,payload,-1);
     }

//------------------------------------------------------------------------------
//  response
//------------------------------------------------------------------------------
public void rsp(SpaceChannel channelElement,String payload,int serial) throws SpaceTimeout,SpaceException {
     //-------------------------------------------------------------------------
     //  build response
     //-------------------------------------------------------------------------
     StringBuffer request = new StringBuffer();

     request.append("RSP ");
     request.append(". ");
     if(serial == -1) {
          int normalserial = serialCache.get();
          request.append(String.valueOf(normalserial));
          }
     else {
          request.append(String.valueOf(serial));
          }
     request.append(" ");
     request.append(String.valueOf(channelElement.getSeqno()));
     request.append(" ");
     request.append(payload.length());
     request.append(" ");
     request.append("+");
     request.append("\r\n");
     String mime = channelElement.getMime();
     if(mime != null) request.append(mime);
     request.append("\r\n");
     request.append(payload);
     request.append("END\r\n");
 
     //-------------------------------------------------------------------------
     //  write request
     //-------------------------------------------------------------------------
     socket.write(request.toString());

     //-------------------------------------------------------------------------
     //  update seqno for the channel
     //-------------------------------------------------------------------------
     channelElement.setNextSeqno(payload.length());

     //-------------------------------------------------------------------------
     //  mark channel associated with the serial available
     //-------------------------------------------------------------------------
     requestQueue.free(serial);

     //-------------------------------------------------------------------------
     //  notify all listeners
     //-------------------------------------------------------------------------
     thread.threadNotifyAll();
     }
 
//------------------------------------------------------------------------------
//  get a request to process
//------------------------------------------------------------------------------
public SpaceRequest getRequest() {
     while(true) {
          //--------------------------------------------------------------------
          //  get request queue item if present
          //--------------------------------------------------------------------
          SpaceRequest request = requestQueue.get();

          if(request != null) {
               //---------------------------------------------------------------
               //  remove request from server data base
               //---------------------------------------------------------------
               free(request,false);

               return(request);
               }

          //--------------------------------------------------------------------
          //  nothing to do
          //--------------------------------------------------------------------
          return(null);
          }
     }

//------------------------------------------------------------------------------
//  wait for any event
//------------------------------------------------------------------------------
public void threadWait(long timeout) {
     try {
          thread.threadWait(timeout);
          }
     catch(InterruptedException e) {
          }
     }

//------------------------------------------------------------------------------
//  wait for complete response
//------------------------------------------------------------------------------
public SpaceResponse wait(SpaceResponse response) throws SpaceException {
     return(wait(response,0));
     }

//------------------------------------------------------------------------------
//  wait for complete response with timeout specified
//------------------------------------------------------------------------------
public SpaceResponse wait(SpaceResponse response,long timeout) throws SpaceException {
     if(thread == null) {
          throw new SpaceException("SpaceServer: Thread Not Running");
          }

     if(response == null) {
          return(null);
          }

     response = thread.wait(response,timeout);

     if(response.isComplete()) {
          free(response,false);
          }

     return(response);
     }

//------------------------------------------------------------------------------
//  response free including flushing response content
//------------------------------------------------------------------------------
public void free(SpacePacket packet) {
     free(packet,true);
     }

//------------------------------------------------------------------------------
//  response free
//------------------------------------------------------------------------------
private void free(SpacePacket packet,boolean freePacket) {
     //-------------------------------------------------------------------------
     //  get serial for the response
     //-------------------------------------------------------------------------
     int serial = packet.getSerial();

     //-------------------------------------------------------------------------
     //  remove response from response cache
     //-------------------------------------------------------------------------
     cache.remove(serial,freePacket);

     //-------------------------------------------------------------------------
     //  serial is now available for other requests
     //-------------------------------------------------------------------------
     serialCache.set(serial);
     }

//------------------------------------------------------------------------------
//  server close
//------------------------------------------------------------------------------
public void close() {
     close(10000); // up to 10 second settling
     }

//------------------------------------------------------------------------------
//  server close
//------------------------------------------------------------------------------
private void close(int delay) {
     //-------------------------------------------------------------------------
     //  end the session
     //-------------------------------------------------------------------------
     try {
          wait(respond(channelZero,"",false),delay);
          }
     catch(SpaceTimeout e) {
          }
     catch(SpaceException e) {
          }
     catch(NullPointerException e) {
          }

     //-------------------------------------------------------------------------
     //  stop the response thread
     //-------------------------------------------------------------------------
     try {
          if(thread != null) thread.threadStop();
          }
     catch(InterruptedException e) {
          }
     catch(NullPointerException e) {
          }

     //-------------------------------------------------------------------------
     //  close the socket
     //-------------------------------------------------------------------------
     try {
          if(socket != null) socket.close();
          }
     catch(NullPointerException e) {
          }
     }

//------------------------------------------------------------------------------
//  get remote host address
//------------------------------------------------------------------------------
public InetAddress getInetAddress() {
     if(socket == null) return(null);
     return(socket.getInetAddress());
     }

public String getRemoteAddress() {
     if(socket == null) return(null);
     InetAddress address = socket.getInetAddress();
     if(address == null) return("");
     return(address.getHostAddress());
     }

//------------------------------------------------------------------------------
//  set timeout
//------------------------------------------------------------------------------
public void setTimeout(int value) throws SpaceException {
     if(socket != null) socket.setTimeout(value);
     }

//------------------------------------------------------------------------------
//  logging with supplied message
//------------------------------------------------------------------------------
public void log(int level,String message) throws SpaceException {
     if(log != null) log.println(level,message,null);
     }   

public void log(int level,String message,int port) throws SpaceException {
     if(log != null) log.println(level,message,port);
     }   

public void log(int level,String message,SpacePacket packet) throws SpaceException {
     if(log != null) log.println(level,message,packet);
     }   

//------------------------------------------------------------------------------
//  logging with localization support
//------------------------------------------------------------------------------
public void log(int level,int messageCode) throws SpaceException {
     if(log != null) log.println(level,getMessage(messageCode),null);
     }   

public void log(int level,int messageCode,int port) throws SpaceException {
     if(log != null) log.println(level,getMessage(messageCode),port);
     }   

public void log(int level,int messageCode,SpacePacket packet) throws SpaceException {
     if(log != null) log.println(level,getMessage(messageCode),packet);
     }   

//------------------------------------------------------------------------------
//  set tracing
//------------------------------------------------------------------------------
public void setTrace(boolean value) {
     if(thread != null) thread.setTrace(value);
     trace = value;
     }

//------------------------------------------------------------------------------
//  trace
//------------------------------------------------------------------------------
public void trace() {
     if(cache != null) cache.trace();
     }

//------------------------------------------------------------------------------
//  set debug
//------------------------------------------------------------------------------
public void setDebug(boolean value) {
     debug = value; 
     if(socket != null) socket.setDebug(value); 
     }

//------------------------------------------------------------------------------
//  get localize (with default)
//------------------------------------------------------------------------------
public String getLocalize() {
     if((localize == null) || (localize.length() == 0)) return("en-us");
     return(localize);
     }

//------------------------------------------------------------------------------
//  localized message
//------------------------------------------------------------------------------
public String getMessage(int messageCode) throws SpaceException {
     InvisibleVector v = InvisibleUtil.parse(getLocalize()," ");
     int size = v.size();
     for(int k=0;k<size;++k) {
          lastLocalize = v.getString(k);
          SpaceMessage message = SpaceMessageLoad.getClass(lastLocalize);
          if(message == null) continue;
          return(message.get(messageCode));
          }

     lastLocalize = "i-default";
     SpaceMessage message = SpaceMessageLoad.getClass(lastLocalize);
     if(message == null) {
          throw new SpaceException("SpaceServer: Unable To Localize");
          }

     return(message.get(messageCode));
     }

//------------------------------------------------------------------------------
//  socket bean property access methods
//------------------------------------------------------------------------------
public void setHost(String value) { socket.setHost(value); }
public void setPort(int value) { socket.setPort(value); }
public void setLog(SpaceLog value) { log = value; }
public void setAuthenticator(String value) { authenticator = value; }

public SpaceChannel getChannelZero() { return(channelZero); }
public SpaceResponse getChannelZeroResponse() { return(channelZeroResponse); }
public String getHost() { return(socket.getHost()); }
public int getPort() { return(socket.getPort()); }
public SpaceLog getLog() { return(log); }
public String[] getProfiles() { return(profiles); }
public String getFeatures() { return(features); }
public String getAuthenticator() { return(authenticator); }
}
