//------------------------------------------------------------------------------
//  Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
// 
//  The contents of this file are subject to the Blocks Public License (the 
//  "License"); you may not use this file except in compliance with the License.  
//  You may obtain a copy of the License at http://www.invisible.net/ 
// 
//  Software distributed under the License is distributed on an "AS IS" basis,  
//  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for  
//  the specific language governing rights and limitations under the License. 
//------------------------------------------------------------------------------
public class SpaceServerThread extends Thread {

private SpaceSocket                     socket;
private SpaceSerialCache                serialCache;
//private SpaceChannel                    channelZero;
private SpaceChannelCache               requestChannels;
private SpaceRequestQueue               requestQueue;
private SpaceCache                      cache;
private SpaceLog                        log;
private boolean                         shutdownOnTimeout;
private boolean                         shutdown;
private boolean                         trace;

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceServerThread(boolean shutdownOnTimeoutArg) throws SpaceException {
     shutdownOnTimeout = shutdownOnTimeoutArg;
     shutdown = false;
     trace = false;
     }

//------------------------------------------------------------------------------
//  runnable response
//------------------------------------------------------------------------------
public void run() {
     while(true) {
          try {
               //---------------------------------------------------------------
               //  response loop--read until timeout
               //---------------------------------------------------------------
               response();

               //---------------------------------------------------------------
               //  see if shutdown
               //---------------------------------------------------------------
               if(shutdown) break;
               }
          catch(SpaceTimeout e) {
               //---------------------------------------------------------------
               //  timeout or shutdown
               //---------------------------------------------------------------
               if(shutdownOnTimeout) break;
               if(shutdown) break;
               }
          catch(SpaceException e) {
               //---------------------------------------------------------------
               //  exception
               //---------------------------------------------------------------
               System.err.println("SpaceServer: " + e.toString());
               break;
               }
          catch(InterruptedException e) {
               //---------------------------------------------------------------
               //  thread exception
               //---------------------------------------------------------------
               System.err.println("SpaceServer: " + e.toString());
               break;
               }
          }
     }

//------------------------------------------------------------------------------
//  wait for complete response
//------------------------------------------------------------------------------
synchronized public SpaceResponse wait(SpaceResponse response) throws SpaceException {
     return(wait(response,0));
     }

//------------------------------------------------------------------------------
//  wait for complete response with timeout specified
//------------------------------------------------------------------------------
synchronized public SpaceResponse wait(SpaceResponse response,long timeout) throws SpaceException {
     boolean nowait = false;

     while(true) {
          //--------------------------------------------------------------------
          //  shutdown
          //--------------------------------------------------------------------
          if(shutdown) {
               response.setError(500,"lost connection");
               return(response);
               }

          //--------------------------------------------------------------------
          //  see if response complete
          //--------------------------------------------------------------------
          if(response.isComplete()) {
               //---------------------------------------------------------------
               //  complete response--trace if appropriate
               //---------------------------------------------------------------
               if(trace) {
                    cache.trace();
                    }

               //---------------------------------------------------------------
               //  return complete response
               //---------------------------------------------------------------
               return(response);
               }

          //--------------------------------------------------------------------
          //  if previous timeout do not wait
          //--------------------------------------------------------------------
          if(nowait) {
               return(response);
               }

          //--------------------------------------------------------------------
          //  wait for notify as specified
          //--------------------------------------------------------------------
          try {
              if(timeout <= 0) {
                    //----------------------------------------------------------
                    //  wait until notified
                    //----------------------------------------------------------
                    wait();
                    }
               else {
                    //----------------------------------------------------------
                    //  wait for specified timeout
                    //----------------------------------------------------------
                    wait(timeout);
                    nowait = true;
                    }
               }
          catch(InterruptedException e) {
               throw new SpaceException("SpaceServer: " + e.toString());
               }
          }
     }
 
//------------------------------------------------------------------------------
//  read response frames
//------------------------------------------------------------------------------
private void response() throws SpaceTimeout,SpaceException,InterruptedException {
     try {
          //--------------------------------------------------------------------
          //  get response
          //--------------------------------------------------------------------
          String line = null;
 
          while((line = socket.read()) != null) {
               //---------------------------------------------------------------
               //  ignore diagnostic messages--remove from production code
               //---------------------------------------------------------------
               if(line.startsWith("DEBUG: ")) {
                    continue;
                    }

               //---------------------------------------------------------------
               //  seq
               //---------------------------------------------------------------
               if(line.startsWith("SEQ")) {
                    continue;
                    }
 
               //---------------------------------------------------------------
               //  req
               //---------------------------------------------------------------
               SpacePacket response = null;
               SpaceHeader header = null;

               if(line.startsWith("REQ")) {
                    //----------------------------------------------------------
                    //  parse request header
                    //----------------------------------------------------------
                    header = new SpaceRequestHeader(line);

                    //----------------------------------------------------------
                    //  get request for this serial if it exists
                    //----------------------------------------------------------
                    response = cache.get(header.getSerial());

                    if(response == null) {
                         //-----------------------------------------------------
                         //  new request--get channel
                         //-----------------------------------------------------
                         int channel = ((SpaceRequestHeader)(header)).getChannel();

                         SpaceChannel channelElement = requestChannels.get(channel);

                         if(channelElement == null) {
                              throw new SpaceException("SpaceServerThread: Channel Not Open: " + channel);
                              }

                         //-----------------------------------------------------
                         //  instantiate the request
                         //-----------------------------------------------------
                         int serial = header.getSerial();
                         response = new SpaceRequest(channelElement,serial);

                         cache.set(serial,response);
                         }
                    }
 
               //---------------------------------------------------------------
               //  rsp
               //---------------------------------------------------------------
               if(line.startsWith("RSP")) {
                    //----------------------------------------------------------
                    //  parse response header
                    //----------------------------------------------------------
                    header = new SpaceResponseHeader(line);

                    //----------------------------------------------------------
                    //  get response for this serial
                    //----------------------------------------------------------
                    response = cache.get(header.getSerial());

                    if(response == null) {
                         //-----------------------------------------------------
                         //  instantiate the request
                         //-----------------------------------------------------
                         throw new SpaceException("SpaceServerThread: Response Missing");
                         }
                    }

               //---------------------------------------------------------------
               //  get stream if valid header has been found
               //---------------------------------------------------------------
               if(header != null) {
                    //----------------------------------------------------------
                    //  get mime type on first frame and increment frame count
                    //----------------------------------------------------------
                    if(response.incrementFrameCount() == 0) {
                         String mime = socket.read();
                         response.getChannelElement().setMime(mime);
                         }
 
                    //----------------------------------------------------------
                    //  get specified number of bytes
                    //----------------------------------------------------------
                    int size = header.getSize();

                    //----------------------------------------------------------
                    //  see if shutdown
                    //----------------------------------------------------------
                    if((response.getChannel() == 0) && (size == 0)) {
                         shutdown = true;
                         }

                    //----------------------------------------------------------
                    //  get the frame
                    //----------------------------------------------------------
                    if(size > 0) {
                         socket.read(response,size);
                         }

                    //----------------------------------------------------------
                    //  should be end if complete response
                    //----------------------------------------------------------
                    String end = socket.read();

                    if(!(end.startsWith("END"))) {
                         throw new SpaceException("SpaceSocket: Response END Expected: " + line);
                         }

                    //----------------------------------------------------------
                    //  save header
                    //----------------------------------------------------------
                    response.setHeader(header);

                    //----------------------------------------------------------
                    //  send seq cleanup unless greeting
                    //----------------------------------------------------------
                    if(!(shutdown)) {
                         seq(response.getChannelElement(),header.getSeqno(),header.getSize());
                         }

                    //----------------------------------------------------------
                    //  if response is complete notify any waiting threads
                    //----------------------------------------------------------
                    if((header.isComplete()) || (header.isError())) {
                         //-----------------------------------------------------
                         //  mark response complete
                         //-----------------------------------------------------
                         response.setIncomplete(false);
                         response.setError(response.isError());

                         //-----------------------------------------------------
                         //  trace on complete response
                         //-----------------------------------------------------
                         if(trace) {
                              cache.trace();
                              }

                         //-----------------------------------------------------
                         //  parse response if error
                         //-----------------------------------------------------
                         if(header.isError()) {
                              response.parse();
                              }

                         //-----------------------------------------------------
                         //  if request put in the request queue
                         //-----------------------------------------------------
                         if(header.isRequest()) {
                              requestQueue.set(response);
                              }

                         //-----------------------------------------------------
                         //  notify all listeners
                         //-----------------------------------------------------
                         threadNotifyAll();
                         }
                    else {
                         //-----------------------------------------------------
                         //  trace on incomplete response
                         //-----------------------------------------------------
                         if(trace) {
                              cache.trace();
                              }
                         }

                    if(shutdown) break;
                    continue;
                    } 

               //---------------------------------------------------------------
               //  unknown response
               //---------------------------------------------------------------
               if(!(shutdown)) {
                    throw new SpaceException("SpaceSocket: Unknown Response: " + line);
                    }
               }

          //--------------------------------------------------------------------
          //  premature eof
          //--------------------------------------------------------------------
          if(!(shutdown)) {
               throw new SpaceException("SpaceServer: Premature EOF");
               }
          }
     catch(SpaceTimeout e) {
          //--------------------------------------------------------------------
          //  timeout--wait for reconnect unless shutdown
          //--------------------------------------------------------------------
          if(shutdownOnTimeout) {
               shutdown = true;
               threadNotifyAll();
               }

          if(!(shutdown)) threadWait();

          throw new SpaceTimeout("SpaceServer: Timeout");
          }
     }

//------------------------------------------------------------------------------
//  thread notify all
//------------------------------------------------------------------------------
synchronized public void threadNotifyAll() {
     notifyAll();
     }

//------------------------------------------------------------------------------
//  thread wait with timeout
//------------------------------------------------------------------------------
synchronized public void threadWait(long timeout) throws InterruptedException {
     wait(timeout);
     }

//------------------------------------------------------------------------------
//  thread wait
//------------------------------------------------------------------------------
synchronized public void threadWait() throws InterruptedException {
     wait();
     }

//------------------------------------------------------------------------------
//  thread stop
//------------------------------------------------------------------------------
public void threadStop() throws InterruptedException {
     shutdown = true;
     threadNotifyAll();
     }

//------------------------------------------------------------------------------
//  seq
//------------------------------------------------------------------------------
private void seq(SpaceChannel channelElement,long seqno,int size) throws SpaceException {
     StringBuffer request = new StringBuffer();

     request.append("SEQ ");
     request.append(String.valueOf(channelElement.getChannel()));
     request.append(" ");
     request.append(seqno);
     request.append(" ");
     request.append((seqno+size+4096));
     request.append("\r\n");

     socket.write(request.toString());
     }   

//------------------------------------------------------------------------------
//  socket bean property access methods
//------------------------------------------------------------------------------
public void setSocket(SpaceSocket value) { socket = value; }
public void setSerialCache(SpaceSerialCache value) { serialCache = value; }
public void setRequestChannels(SpaceChannelCache value) { requestChannels = value; }
public void setRequestQueue(SpaceRequestQueue value) { requestQueue = value; }
public void setCache(SpaceCache value) { cache = value; }
public void setLog(SpaceLog value) { log = value; }
public void setShutdown(boolean value) { shutdown = value; }
public void setTrace(boolean value) { trace = value; }
}
