//------------------------------------------------------------------------------
//  Copyright (c) 2000, Invisible Worlds, Inc. All rights reserved.
// 
//  The contents of this file are subject to the Blocks Public License (the 
//  "License"); you may not use this file except in compliance with the License.  
//  You may obtain a copy of the License at http://www.invisible.net/ 
// 
//  Software distributed under the License is distributed on an "AS IS" basis,  
//  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for  
//  the specific language governing rights and limitations under the License. 
//------------------------------------------------------------------------------
import java.net.ServerSocket;
import java.net.*;
import java.io.*;

public class SpaceSocket {

public final static int                 defaultTimeout = 600000; //10 minutes

private String                          host;
private int                             port;
private int                             timeout;

private InetAddress                     address;
private Socket                          hostSocket;
private ServerSocket                    serverSocket;
private DataOutputStream                hostOutput;
//private BufferedWriter                  hostOutput;
private BufferedReader                  hostInput;
private boolean                         debug;

//------------------------------------------------------------------------------
//  constructor
//------------------------------------------------------------------------------
public SpaceSocket(int portArg) {
     port = portArg;
     timeout = defaultTimeout;

     hostSocket = null;
     hostOutput = null;
     hostInput = null;
     }

//------------------------------------------------------------------------------
//  connect
//------------------------------------------------------------------------------
public void connect() throws SpaceException {
     try {
          //--------------------------------------------------------------------
          //  instantiate the socket and i/o streams
          //--------------------------------------------------------------------
          address=InetAddress.getByName(host);
          hostSocket=new Socket(address,port);
          setStreams();
          }
     catch(IOException e) {
          close();
          throw new SpaceException("SpaceSocket: " + e.toString());
          }
     }

//------------------------------------------------------------------------------
//  accept
//------------------------------------------------------------------------------
public void accept(boolean useStdio) throws SpaceException {
     //-------------------------------------------------------------------------
     //  see if accepting inetd style connection
     //-------------------------------------------------------------------------
     if(useStdio) {
          //--------------------------------------------------------------------
          //  inetd style connection
          //--------------------------------------------------------------------
          hostOutput = new DataOutputStream(System.out);
          hostInput = new BufferedReader(new InputStreamReader(System.in));
          return;
          }

     //-------------------------------------------------------------------------
     //  wait for a connection on the specified port
     //-------------------------------------------------------------------------
     try {
          serverSocket = new ServerSocket(port);
          hostSocket = serverSocket.accept();
          setStreams();
          }
     catch(IOException e) {
          close();
          throw new SpaceException("SpaceSocket: " + e.toString());
          }
     }

//------------------------------------------------------------------------------
//  write
//------------------------------------------------------------------------------
synchronized public void write(String string) throws SpaceException {
     if(hostOutput == null) {
          throw new SpaceException("SpaceSocket: Output To Null Socket");
          }

     try {
          hostOutput.writeBytes(string);
          //hostOutput.write(string,0,string.length());
          if(debug) System.err.println("Write: " + string);
          }
     catch(IOException e) {
          close();
          throw new SpaceException("SpaceSocket: " + e.toString());
          }
     }

//------------------------------------------------------------------------------
//  read
//------------------------------------------------------------------------------
public String read() throws SpaceTimeout,SpaceException {
     try {
          String line = hostInput.readLine();

          if(line == null) {
               close();
               throw new SpaceTimeout("SpaceSocket: Timeout");
               }

          if(debug) System.err.println("Read: " + line);

          return(line);
          }
     catch(NullPointerException e) {
          //--------------------------------------------------------------------
          //  null exception generally means shutdown in progress
          //--------------------------------------------------------------------
          close();
          throw new SpaceTimeout("SpaceSocket: Timeout");
          }
     catch(InterruptedIOException e) {
          close();
          throw new SpaceTimeout("SpaceSocket: Timeout");
          }
     catch(java.net.SocketException e) {
          close();
          throw new SpaceTimeout("SpaceSocket: Timeout");
          }
     catch(IOException e) {
          close();
          throw new SpaceException("SpaceSocket: " + e.toString());
          }
     }

//------------------------------------------------------------------------------
//  read specified number of octets
//------------------------------------------------------------------------------
public int read(SpacePacket response,int size) throws SpaceTimeout,SpaceException {
     if(hostInput == null) {
          throw new SpaceException("SpaceSocket: Input From Null Socket");
          }

     try {
          int inchar;

          for(int k=0;k<size;++k) {
               if((inchar = hostInput.read()) == -1) {
                    close();
                    throw new SpaceTimeout("SpaceSocket: Timeout");
                    }
               response.append((char)(inchar));
               }

          if(debug) System.err.println("Read: " + response.toString());

          return(size);
          }
     catch(java.io.InterruptedIOException e) {
          close();
          throw new SpaceTimeout("SpaceSocket: Timeout");
          }
     catch(java.net.SocketException e) {
          close();
          throw new SpaceTimeout("SpaceSocket: Timeout");
          }
     catch(IOException e) {
          close();
          throw new SpaceException("SpaceSocket: " + e.toString());
          }
     }

//------------------------------------------------------------------------------
//  close
//------------------------------------------------------------------------------
public void close() {
     try {
          if(hostOutput != null) {
               hostOutput.flush();
               hostOutput.close();
               hostOutput = null;
               }
          }
     catch(IOException e) {
          hostOutput = null;
          }

     try {
          if(hostInput != null) {
               while(hostInput.ready()) hostInput.read();
               hostInput.close();
               hostInput = null;
               }
          }
     catch(IOException e) {
          hostInput = null;
          }

     try {
          if(hostSocket != null) {
               hostSocket.close();
               hostSocket = null;
               }
          }
     catch(IOException e) {
          hostSocket = null;
          }

     try {
          if(serverSocket != null) {
               serverSocket.close();
               serverSocket = null;
               }
          }
     catch(IOException e) {
          serverSocket = null;
          }
     }

//------------------------------------------------------------------------------
//  finalize
//------------------------------------------------------------------------------
public void finalize() throws IOException {
     close();
     }

//------------------------------------------------------------------------------
//  set timeout
//------------------------------------------------------------------------------
public void setTimeout(int value) throws SpaceException {
     timeout = value;
     if(hostSocket != null) {
          try {
               hostSocket.setSoTimeout(timeout);
               }
          catch(java.net.SocketException e) {
               close();
               throw new SpaceException("SpaceSocket: Unable To Set Timeout");
               }
          }
     }

//------------------------------------------------------------------------------
//  set i/o streams
//------------------------------------------------------------------------------
private void setStreams() throws SpaceException,IOException {
     setTimeout(timeout);
     hostOutput=new DataOutputStream(hostSocket.getOutputStream());
     hostInput = new BufferedReader(new InputStreamReader(hostSocket.getInputStream()));
     }

//------------------------------------------------------------------------------
//  get address of remote machine
//------------------------------------------------------------------------------
public InetAddress getInetAddress() {
     if(hostSocket == null) return(null);
     return(hostSocket.getInetAddress());
     }

//------------------------------------------------------------------------------
//  bean property access methods
//------------------------------------------------------------------------------
public void setHost(String value) { host = value; }
public void setPort(int value) { port = value; }
public void setDebug(boolean value) { debug = value; }

public String getHost() { return(host); }
public int getPort() { return(port); }
public boolean getDebug() { return(debug); }
}
